from __future__ import print_function

import sys

import pytest

from .util import silence

try:
    from configparser import ConfigParser
except ImportError:
    from ConfigParser import ConfigParser

from .constants import EXIT_INTERRUPTED, EXIT_NOTESTSCOLLECTED, EXIT_OK


CLI_OPTION_PREFIX = '--'


class CollectError(Exception):
    pass


class StopCollect(Exception):
    pass


class CollectConfig(object):
    """
    A pytest plugin to gets the configuration file.
    """
    def __init__(self):
        self.path = None

    def pytest_cmdline_main(self, config):
        if hasattr(config, 'inifile'):
            # pytest >= 2.7.0
            inifile = config.inifile
        else:
            # pytest < 2.7.0
            inifile = config.inicfg.config.path
        if inifile:
            self.path = str(inifile)
            raise StopCollect()


def _run_pytest_collect(pytest_args):
    collect_config_plugin = CollectConfig()
    argv = pytest_args + ['--collect-only']

    try:
        exit_code = pytest.main(argv, plugins=[collect_config_plugin])
    except StopCollect:
        return collect_config_plugin.path

    if exit_code == EXIT_INTERRUPTED:
        raise KeyboardInterrupt()
    if exit_code not in [EXIT_OK, EXIT_NOTESTSCOLLECTED]:
        raise CollectError()

    return collect_config_plugin.path


def _collect_config(pytest_args, silent=True):
    if silent:
        try:
            with silence():
                return _run_pytest_collect(pytest_args)
        except KeyboardInterrupt:
            raise
        except Exception:
            pass
        # Print message and run again without silencing
        print('Error: Could not run --collect-only to find the pytest config '
              'file. Trying again without silencing stdout...',
              file=sys.stderr)

    return _run_pytest_collect(pytest_args)


def merge_config(args, pytest_args, silent=True, verbose=False):
    if verbose:
        print('Locating inifile...')

    try:
        config_path = _collect_config(pytest_args, silent)
    except (KeyboardInterrupt, CollectError):
        return False

    if not config_path:
        return True

    config = ConfigParser()
    config.read(config_path)
    if not config.has_section('pytest-watch'):
        return True

    for cli_name in args:
        if not cli_name.startswith(CLI_OPTION_PREFIX):
            continue
        config_name = cli_name[len(CLI_OPTION_PREFIX):]

        # Let CLI options take precedence
        if args[cli_name]:
            continue

        # Find config option
        if not config.has_option('pytest-watch', config_name):
            continue

        # Merge config option using the expected type
        if isinstance(args[cli_name], list):
            args[cli_name].append(config.get('pytest-watch', config_name))
        elif isinstance(args[cli_name], bool):
            args[cli_name] = config.getboolean('pytest-watch', config_name)
        else:
            args[cli_name] = config.get('pytest-watch', config_name)

    return True



"""
pytest_watch.command
~~~~~~~~~~~~~~~~~~~~

Implements the command-line interface for pytest-watch.


Usage: ptw [options] [--ignore <dir>...] [<directory>...] [-- <pytest-args>...]

Options:
  --ignore <dir>        Ignore directory from being watched and during
                        collection (multi-allowed).
  --ext <exts>          Comma-separated list of file extensions that can
                        trigger a new test run when changed (default: .py).
                        Use --ext=* to allow any file (including .pyc).
  --config <file>       Load configuration from `file` instead of trying to
                        locate one of the implicit configuration files.
  -c --clear            Clear the screen before each run.
  -n --nobeep           Do not beep on failure.
  -w --wait             Waits for all tests to complete before re-running.
                        Otherwise, tests are interrupted on filesystem events.
  --beforerun <cmd>     Run arbitrary command before tests are run.
  --afterrun <cmd>      Run arbitrary command on completion or interruption.
                        The exit code of "py.test" is passed as an argument.
  --onpass <cmd>        Run arbitrary command on pass.
  --onfail <cmd>        Run arbitrary command on failure.
  --onexit <cmd>        Run arbitrary command when exiting pytest-watch.
  --runner <cmd>        Run a custom command instead of "py.test".
  --pdb                 Start the interactive Python debugger on errors.
                        This also enables --wait to prevent pdb interruption.
  --spool <delay>       Re-run after a delay (in milliseconds), allowing for
                        more file system events to queue up (default: 200 ms).
  -p --poll             Use polling instead of OS events (useful in VMs).
  -v --verbose          Increase verbosity of the output.
  -q --quiet            Decrease verbosity of the output (precedence over -v).
  -V --version          Print version and exit.
  -h --help             Print help and exit.
"""

import sys

import colorama
from docopt import docopt

from . import __version__
from .config import merge_config
from .constants import ALL_EXTENSIONS
from .watcher import watch


doc = '\n\n\n'.join(__doc__.split('\n\n\n')[1:])
version = 'pytest-watch ' + __version__


def main(argv=None):
    """
    The entry point of the application.
    """
    if argv is None:
        argv = sys.argv[1:]

    # Initialize terminal colors
    colorama.init()

    # Parse CLI arguments
    args = docopt(doc, argv=argv, version=version)

    # Get paths and initial pytest arguments
    directories = args['<directory>']
    pytest_args = list(directories)
    if '--' in directories:
        index = directories.index('--')
        directories = directories[:index]
        del pytest_args[index]

    # Adjust pytest and --collect-only args
    for ignore in args['--ignore']:
        pytest_args.extend(['--ignore', ignore])
    if args['--config']:
        pytest_args.extend(['-c', args['--config']])

    # Merge config file options
    if not merge_config(args, pytest_args, verbose=args['--verbose']):
        return 0

    # Adjust pytest args
    if args['--pdb']:
        pytest_args.append('--pdb')

    # Parse extensions
    if args['--ext'] == '*':
        extensions = ALL_EXTENSIONS
    elif args['--ext']:
        extensions = [('.' if not e.startswith('.') else '') + e
                      for e in args['--ext'].split(',')]
    else:
        extensions = None

    # Parse numeric arguments
    spool = args['--spool']
    if spool is not None:
        try:
            spool = int(spool)
        except ValueError:
            sys.stderr.write('Error: Spool must be an integer.\n')
            return 2

    # Run pytest and watch for changes
    return watch(directories=directories,
                 ignore=args['--ignore'],
                 extensions=extensions,
                 beep_on_failure=not args['--nobeep'],
                 auto_clear=args['--clear'],
                 wait=args['--wait'] or args['--pdb'],
                 beforerun=args['--beforerun'],
                 afterrun=args['--afterrun'],
                 onpass=args['--onpass'],
                 onfail=args['--onfail'],
                 onexit=args['--onexit'],
                 runner=args['--runner'],
                 spool=spool,
                 poll=args['--poll'],
                 verbose=args['--verbose'],
                 quiet=args['--quiet'],
                 pytest_args=pytest_args)



"""
pytest_watch
------------

Local continuous test runner with pytest and watchdog.

:copyright: (c) 2014 by Joe Esposito.
:license: MIT, see LICENSE for more details.
"""

import os
import sys


if __name__ == '__main__':
    sys.path.append(os.path.dirname(__file__))

    from pytest_watch.command import main
    main()



import os
import sys
from contextlib import contextmanager


@contextmanager
def silence():
    """
    Silence stdout and stderr in a 'with' block.
    """
    old_stdout = sys.stdout
    old_stderr = sys.stderr

    nullfd = open(os.devnull, 'w')
    sys.stdout = nullfd
    sys.stderr = nullfd

    try:
        yield
    except Exception:
        raise
    finally:
        sys.stdout = old_stdout
        sys.stderr = old_stderr
        nullfd.close()



import os
import signal
import subprocess
import sys
from time import sleep

try:
    from queue import Empty
except ImportError:
    from Queue import Empty


is_windows = sys.platform == 'win32'


def beep():
    """
    Emits a beep sound.
    """
    sys.stdout.write('\a')
    sys.stdout.flush()


def clear():
    """
    Clears the terminal.
    """
    subprocess.call('cls' if is_windows else 'clear', shell=True)


def dequeue_all(queue, spool=None):
    """
    Empties the specified queue into a list, optionally with spooling.

    Spool default is 200 (ms). Set to 0 to disable.
    """
    if spool is None:
        spool = 200
    items = []
    while True:
        try:
            while True:
                items.append(queue.get_nowait())
        except Empty:
            # If spooling, wait a moment and check for new items
            if spool:
                sleep(spool / 1000.0)
                if not queue.empty():
                    continue
            break
    return items


def samepath(left, right):
    """
    Determines whether two paths are the same.
    """
    return (os.path.abspath(os.path.normcase(left)) ==
            os.path.abspath(os.path.normcase(right)))


def send_keyboard_interrupt(proc):
    """
    Sends a KeyboardInterrupt to the specified child process.
    """
    if is_windows:
        try:
            # Send KeyboardInterrupt to self, and therefore, to child processes
            try:
                os.kill(0, signal.CTRL_C_EVENT)
            except AttributeError:
                # Python 2.6 and below
                import ctypes
                ctypes.windll.kernel32.GenerateConsoleCtrlEvent(0, 0)
            # Immediately throws KeyboardInterrupt from the simulated CTRL-C
            proc.wait()
        except KeyboardInterrupt:
            # Ignore the simulated CTRL-C
            pass
    else:
        os.kill(proc.pid, signal.SIGINT)



"""
pytest_watch
------------

Local continuous test runner with pytest and watchdog.

:copyright: (c) 2014-2016 by Joe Esposito.
:license: MIT, see LICENSE for more details.
"""

__version__ = '4.1.0'


from .command import main, doc, version
from .watcher import watch


__all__ = ['main', 'doc', 'version', 'watch']
